

/*
 * Author: Tomi Jylhä-Ollila, Finland 2010
 *
 * This file is part of Kunquat.
 *
 * CC0 1.0 Universal, http://creativecommons.org/publicdomain/zero/1.0/
 *
 * To the extent possible under law, Kunquat Affirmers have waived all
 * copyright and related or neighboring rights to Kunquat.
 */


#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

#include <Voice_pool.h>
#include <xassert.h>
#include <xmemory.h>


#define HEAP_PARENT(i) (((i) - 1) / 2)
#define HEAP_LEFT(i)   (2 * (i) + 1)
#define HEAP_RIGHT(i)  (2 * (i) + 2)


/**
 * Moves a modified key into its proper position in the heap.
 * 
 * The tree must satisfy the minimum heap property apart from the given index.
 *
 * \param pool    The Voice pool -- must not be \c NULL.
 * \param index   The index of the key -- must be less than the pool size.
 *
 * \return   The target index of the modified key.
 */
static uint16_t heap_mod_key(Voice_pool* pool, uint16_t index);


/**
 * Arranges the Voice pool to satisfy the minimum heap property.
 *
 * \param pool   The Voice pool -- must not be \c NULL.
 */
static void heap_build(Voice_pool* pool);


/**
 * Performs a downheap operation starting at specified index.
 *
 * The subtrees must satisfy the minimum heap property.
 *
 * \param pool    The Voice pool -- must not be \c NULL.
 * \param index   The index of the key -- must be less than the pool size.
 *
 * \return   The target index of the modified key.
 */
static uint16_t downheap(Voice_pool* pool, uint16_t index);


/**
 * Performs an upheap operation starting at specified index.
 *
 * The tree must satisfy the minimum heap property apart from the given index.
 *
 * \param pool    The Voice pool -- must not be \c NULL.
 * \param index   The index of the key -- must be less than the pool size.
 *
 * \return   The target index of the modified key.
 */
static uint16_t upheap(Voice_pool* pool, uint16_t index);


Voice_pool* new_Voice_pool(uint16_t size, uint8_t events)
{
    assert(size > 0);
    assert(events > 0);
    Voice_pool* pool = xalloc(Voice_pool);
    if (pool == NULL)
    {
        return NULL;
    }
    pool->size = size;
    pool->state_size = 0;
    pool->events = events;
    pool->voices = xnalloc(Voice, size);
    if (pool->voices == NULL)
    {
        xfree(pool);
        return NULL;
    }
    for (int i = 0; i < size; ++i)
    {
        pool->voices[i] = new_Voice();
        if (pool->voices[i] == NULL)
        {
            for (--i; i >= 0; --i)
            {
                del_Voice(pool->voices[i]);
            }
            xfree(pool->voices);
            xfree(pool);
            return NULL;
        }
    }
    return pool;
}


bool Voice_pool_reserve_state_space(Voice_pool* pool, size_t state_size)
{
    assert(pool != NULL);
    if (state_size <= pool->state_size)
    {
        return true;
    }
    for (uint16_t i = 0; i < pool->size; ++i)
    {
        if (!Voice_reserve_state_space(pool->voices[i], state_size))
        {
            return false;
        }
    }
    pool->state_size = state_size;
    return true;
}


bool Voice_pool_resize(Voice_pool* pool, uint16_t size)
{
    assert(pool != NULL);
    assert(size > 0);
    uint16_t new_size = size;
    if (new_size == pool->size)
    {
        return true;
    }
    for (uint16_t i = new_size; i < pool->size; ++i)
    {
        del_Voice(pool->voices[i]);
        pool->voices[i] = NULL;
    }
    if (new_size < pool->size)
    {
        pool->size = new_size;
    }
    Voice** new_voices = xrealloc(Voice*, new_size, pool->voices);
    if (new_voices == NULL)
    {
        return false;
    }
    pool->voices = new_voices;
    for (uint16_t i = pool->size; i < new_size; ++i)
    {
        pool->voices[i] = NULL;
    }
    for (uint16_t i = pool->size; i < new_size; ++i)
    {
        pool->voices[i] = new_Voice();
        if (pool->voices[i] == NULL ||
                !Voice_reserve_state_space(pool->voices[i], pool->state_size))
        {
            for (--i; i >= pool->size; --i)
            {
                del_Voice(pool->voices[i]);
                pool->voices[i] = NULL;
            }
            return false;
        }
    }
    pool->size = new_size;
    return true;
}


uint16_t Voice_pool_get_size(Voice_pool* pool)
{
    assert(pool != NULL);
    return pool->size;
}


Voice* Voice_pool_get_voice(Voice_pool* pool,
                            Voice* voice,
                            uint64_t id)
{
    assert(pool != NULL);
    if (voice == NULL)
    {
        static uint64_t running_id = 1;
        downheap(pool, 0);
        Voice* new_voice = pool->voices[0];
        new_voice->id = running_id;
        new_voice->prio = VOICE_PRIO_INACTIVE;
        ++running_id;
        return new_voice;
    }
    if (voice->id == id)
    {
        return voice;
    }
    return NULL;
}


void Voice_pool_prepare(Voice_pool* pool)
{
    assert(pool != NULL);
    for (uint16_t i = 0; i < pool->size; ++i)
    {
        if (pool->voices[i]->prio != VOICE_PRIO_INACTIVE)
        {
            Voice_prepare(pool->voices[i]);
        }
    }
    return;
}


uint16_t Voice_pool_mix_bg(Voice_pool* pool,
                           uint32_t amount,
                           uint32_t offset,
                           uint32_t freq,
                           double tempo)
{
    assert(pool != NULL);
    assert(freq > 0);
    uint16_t active_voices = 0;
    for (uint16_t i = 0; i < pool->size; ++i)
    {
        if (pool->voices[i]->prio != VOICE_PRIO_INACTIVE)
        {
            if (pool->voices[i]->prio <= VOICE_PRIO_BG)
            {
//                fprintf(stderr, "Background mix start\n");
                Voice_mix(pool->voices[i], amount, offset, freq, tempo);
//                fprintf(stderr, "Background mix end\n");
            }
            ++active_voices;
        }
    }
    heap_build(pool);
    return active_voices;
}


uint16_t Voice_pool_mix(Voice_pool* pool,
                        uint32_t amount,
                        uint32_t offset,
                        uint32_t freq,
                        double tempo)
{
    assert(pool != NULL);
    assert(freq > 0);
    uint16_t active_voices = 0;
    for (uint16_t i = 0; i < pool->size; ++i)
    {
        if (pool->voices[i]->prio != VOICE_PRIO_INACTIVE)
        {
            Voice_mix(pool->voices[i], amount, offset, freq, tempo);
            ++active_voices;
        }
    }
    heap_build(pool);
    return active_voices;
}


void Voice_pool_fix_priority(Voice_pool* pool, Voice* voice)
{
    assert(pool != NULL);
    assert(voice != NULL);
    assert(pool->voices[voice->pool_index] == voice);
    heap_mod_key(pool, voice->pool_index);
    return;
}


void Voice_pool_reset(Voice_pool* pool)
{
    for (uint16_t i = 0; i < pool->size; ++i)
    {
        Voice_reset(pool->voices[i]);
    }
    return;
}


void del_Voice_pool(Voice_pool* pool)
{
    if (pool == NULL)
    {
        return;
    }
    for (uint16_t i = 0; i < pool->size; ++i)
    {
        del_Voice(pool->voices[i]);
        pool->voices[i] = NULL;
    }
    xfree(pool->voices);
    xfree(pool);
    return;
}


static uint16_t heap_mod_key(Voice_pool* pool, uint16_t index)
{
    assert(pool != NULL);
    assert(index < pool->size);
    index = upheap(pool, index);
    return downheap(pool, index);
}


static void heap_build(Voice_pool* pool)
{
    assert(pool != NULL);
    uint16_t index = HEAP_PARENT(pool->size + 1);
    while (index-- > 0)
    {
        downheap(pool, index);
    }
    return;
}


static uint16_t downheap(Voice_pool* pool, uint16_t index)
{
    assert(pool != NULL);
    assert(index < pool->size);
    uint16_t smallest = index;
    uint32_t left = HEAP_LEFT(index);
    while (left < pool->size)
    {
        if (Voice_cmp(pool->voices[left], pool->voices[smallest]) < 0)
        {
            smallest = left;
        }
        uint32_t right = HEAP_RIGHT(index);
        if (right < pool->size && Voice_cmp(pool->voices[right], pool->voices[smallest]) < 0)
        {
            smallest = right;
        }
        if (index == smallest)
        {
            return index;
        }
        Voice* tmp = pool->voices[index];
        pool->voices[index] = pool->voices[smallest];
        pool->voices[smallest] = tmp;
        pool->voices[index]->pool_index = index;
        pool->voices[smallest]->pool_index = smallest;
        index = smallest;
        left = HEAP_LEFT(index);
    }
    return index;
}


static uint16_t upheap(Voice_pool* pool, uint16_t index)
{
    assert(pool != NULL);
    assert(index < pool->size);
    uint16_t parent = HEAP_PARENT(index);
    while (index > 0 && Voice_cmp(pool->voices[index], pool->voices[parent]) < 0)
    {
        assert(parent < index);
        Voice* tmp = pool->voices[index];
        pool->voices[index] = pool->voices[parent];
        pool->voices[parent] = tmp;
        pool->voices[index]->pool_index = index;
        pool->voices[parent]->pool_index = parent;
        index = parent;
        parent = HEAP_PARENT(index);
    }
    return index;
}


